@@ -1,11 +1,13 @@
 // Provides some very limited but fast packet filter mechanisms
 
-#ifndef PACKETFILTER_H
-#define PACKETFILTER_H
+#pragma once
 
-#include "IP.h"
+#include "IPAddr.h"
 #include "PrefixTable.h"
 
+class IP_Hdr;
+class Val;
+
 class PacketFilter {
 public:
 	explicit PacketFilter(bool arg_default);
@@ -14,10 +16,10 @@ public:
 	// Drops all packets from a particular source (which may be given
 	// as an AddrVal or a SubnetVal) which hasn't any of TCP flags set
 	// (TH_*) with the given probability (from 0..MAX_PROB).
-	void AddSrc(const IPAddr& src, uint32 tcp_flags, double probability);
-	void AddSrc(Val* src, uint32 tcp_flags, double probability);
-	void AddDst(const IPAddr& src, uint32 tcp_flags, double probability);
-	void AddDst(Val* src, uint32 tcp_flags, double probability);
+	void AddSrc(const IPAddr& src, uint32_t tcp_flags, double probability);
+	void AddSrc(Val* src, uint32_t tcp_flags, double probability);
+	void AddDst(const IPAddr& src, uint32_t tcp_flags, double probability);
+	void AddDst(Val* src, uint32_t tcp_flags, double probability);
 
 	// Removes the filter entry for the given src/dst
 	// Returns false if filter doesn not exist.
@@ -31,8 +33,8 @@ public:
 
 private:
 	struct Filter {
-		uint32 tcp_flags;
-		uint32 probability;
+		uint32_t tcp_flags;
+		uint32_t probability;
 	};
 
 	static void DeleteFilter(void* data);
@@ -43,5 +45,3 @@ private:
 	PrefixTable src_filter;
 	PrefixTable dst_filter;
 };
-
-#endif
