@@ -2,6 +2,7 @@
 
 
 #include "zeek-config.h"
+#include "Sessions.h"
 
 #include <netinet/in.h>
 #include <arpa/inet.h>
@@ -9,11 +10,11 @@
 #include <stdlib.h>
 #include <unistd.h>
 
+#include "Desc.h"
 #include "Net.h"
 #include "Event.h"
 #include "Timer.h"
 #include "NetVar.h"
-#include "Sessions.h"
 #include "Reporter.h"
 
 #include "analyzer/protocol/icmp/ICMP.h"
@@ -29,6 +30,8 @@
 #include "TunnelEncapsulation.h"
 
 #include "analyzer/Manager.h"
+#include "iosource/IOSource.h"
+#include "iosource/PktDumper.h"
 
 // These represent NetBIOS services on ephemeral ports.  They're numbered
 // so that we can use a single int to hold either an actual TCP/UDP server
@@ -40,32 +43,6 @@ enum NetBIOS_Service {
 
 NetSessions* sessions;
 
-void TimerMgrExpireTimer::Dispatch(double t, int is_expire)
-	{
-	if ( mgr->LastAdvance() + timer_mgr_inactivity_timeout < timer_mgr->Time() )
-		{
-		// Expired.
-		DBG_LOG(DBG_TM, "TimeMgr %p has timed out", mgr);
-		mgr->Expire();
-
-		// Make sure events are executed.  They depend on the TimerMgr.
-		::mgr.Drain();
-
-		sessions->timer_mgrs.erase(mgr->GetTag());
-		delete mgr;
-		}
-	else
-		{
-		// Reinstall timer.
-		if ( ! is_expire )
-			{
-			double n = mgr->LastAdvance() +
-					timer_mgr_inactivity_timeout;
-			timer_mgr->Add(new TimerMgrExpireTimer(n, mgr));
-			}
-		}
-	}
-
 void IPTunnelTimer::Dispatch(double t, int is_expire)
 	{
 	NetSessions::IPTunnelMap::const_iterator it =
@@ -88,20 +65,6 @@ void IPTunnelTimer::Dispatch(double t, int is_expire)
 
 NetSessions::NetSessions()
 	{
-	TypeList* t = new TypeList();
-	t->Append(base_type(TYPE_ADDR));	// source IP address
-	t->Append(base_type(TYPE_ADDR));	// dest IP address
-	t->Append(base_type(TYPE_COUNT));	// source and dest ports
-
-	ch = new CompositeHash(t);
-
-	Unref(t);
-
-	tcp_conns.SetDeleteFunc(bro_obj_delete_func);
-	udp_conns.SetDeleteFunc(bro_obj_delete_func);
-	icmp_conns.SetDeleteFunc(bro_obj_delete_func);
-	fragments.SetDeleteFunc(bro_obj_delete_func);
-
 	if ( stp_correlate_pair )
 		stp_manager = new analyzer::stepping_stone::SteppingStoneManager();
 	else
@@ -129,16 +92,26 @@ NetSessions::NetSessions()
 		arp_analyzer = new analyzer::arp::ARP_Analyzer();
 	else
 		arp_analyzer = 0;
+
+	memset(&stats, 0, sizeof(SessionStats));
 	}
 
 NetSessions::~NetSessions()
 	{
-	delete ch;
 	delete packet_filter;
 	delete pkt_profiler;
 	Unref(arp_analyzer);
 	delete discarder;
 	delete stp_manager;
+
+	for ( const auto& entry : tcp_conns )
+		Unref(entry.second);
+	for ( const auto& entry : udp_conns )
+		Unref(entry.second);
+	for ( const auto& entry : icmp_conns )
+		Unref(entry.second);
+	for ( const auto& entry : fragments )
+		Unref(entry.second);
 	}
 
 void NetSessions::Done()
@@ -168,7 +141,7 @@ void NetSessions::NextPacket(double t, const Packet* pkt)
 		return;
 		}
 
-	uint32 caplen = pkt->cap_len - pkt->hdr_size;
+	uint32_t caplen = pkt->cap_len - pkt->hdr_size;
 
 	if ( pkt->l3_proto == L3_IPV4 )
 		{
@@ -212,60 +185,7 @@ void NetSessions::NextPacket(double t, const Packet* pkt)
 		DumpPacket(pkt);
 	}
 
-int NetSessions::CheckConnectionTag(Connection* conn)
-	{
-	if ( current_iosrc->GetCurrentTag() )
-		{
-		// Packet is tagged.
-		if ( conn->GetTimerMgr() == timer_mgr )
-			{
-			// Connection uses global timer queue.  But the
-			// packet has a tag that means we got it externally,
-			// probably from the Time Machine.
-			DBG_LOG(DBG_TM, "got packet with tag %s for already"
-					"known connection, reinstantiating",
-					current_iosrc->GetCurrentTag()->c_str());
-			return 0;
-			}
-		else
-			{
-			// Connection uses local timer queue.
-			TimerMgrMap::iterator i =
-				timer_mgrs.find(*current_iosrc->GetCurrentTag());
-			if ( i != timer_mgrs.end() &&
-			     conn->GetTimerMgr() != i->second )
-				{
-				// Connection uses different local queue
-				// than the tag for the current packet
-				// indicates.
-				//
-				// This can happen due to:
-				//     (1) getting same packets with
-				//		different tags
-				//     (2) timer mgr having already expired
-				DBG_LOG(DBG_TM, "packet ignored due old/inconsistent tag");
-				return -1;
-				}
-
-			return 1;
-			}
-		}
-
-	// Packet is not tagged.
-	if ( conn->GetTimerMgr() != timer_mgr )
-		{
-		// Connection does not use the global timer queue.  That
-		// means that this is a live packet belonging to a
-		// connection for which we have already switched to
-		// processing external input.
-		DBG_LOG(DBG_TM, "packet ignored due to processing it in external data");
-		return -1;
-		}
-
-	return 1;
-	}
-
-static unsigned int gre_header_len(uint16 flags)
+static unsigned int gre_header_len(uint16_t flags)
 	{
 	unsigned int len = 4;  // Always has 2 byte flags and 2 byte protocol type.
 
@@ -293,10 +213,10 @@ static unsigned int gre_header_len(uint16 flags)
 void NetSessions::DoNextPacket(double t, const Packet* pkt, const IP_Hdr* ip_hdr,
 			       const EncapsulationStack* encapsulation)
 	{
-	uint32 caplen = pkt->cap_len - pkt->hdr_size;
+	uint32_t caplen = pkt->cap_len - pkt->hdr_size;
 	const struct ip* ip4 = ip_hdr->IP4_Hdr();
 
-	uint32 len = ip_hdr->TotalLen();
+	uint32_t len = ip_hdr->TotalLen();
 	if ( len == 0 )
 		{
 		// TCP segmentation offloading can zero out the ip_len field.
@@ -314,7 +234,7 @@ void NetSessions::DoNextPacket(double t, const Packet* pkt, const IP_Hdr* ip_hdr
 
 	// For both of these it is safe to pass ip_hdr because the presence
 	// is guaranteed for the functions that pass data to us.
-	uint16 ip_hdr_len = ip_hdr->HdrLen();
+	uint16_t ip_hdr_len = ip_hdr->HdrLen();
 	if ( ip_hdr_len > len )
 		{
 		Weird("invalid_IP_header_size", ip_hdr, encapsulation);
@@ -428,7 +348,7 @@ void NetSessions::DoNextPacket(double t, const Packet* pkt, const IP_Hdr* ip_hdr
 	ConnID id;
 	id.src_addr = ip_hdr->SrcAddr();
 	id.dst_addr = ip_hdr->DstAddr();
-	Dictionary* d = 0;
+	ConnectionMap* d = nullptr;
 	BifEnum::Tunnel::Type tunnel_type = BifEnum::Tunnel::IP;
 
 	switch ( proto ) {
@@ -492,8 +412,8 @@ void NetSessions::DoNextPacket(double t, const Packet* pkt, const IP_Hdr* ip_hdr
 			return;
 			}
 
-		uint16 flags_ver = ntohs(*((uint16*)(data + 0)));
-		uint16 proto_typ = ntohs(*((uint16*)(data + 2)));
+		uint16_t flags_ver = ntohs(*((uint16_t*)(data + 0)));
+		uint16_t proto_typ = ntohs(*((uint16_t*)(data + 2)));
 		int gre_version = flags_ver & 0x0007;
 
 		// If a carried packet has ethernet, this will help skip it.
@@ -517,7 +437,7 @@ void NetSessions::DoNextPacket(double t, const Packet* pkt, const IP_Hdr* ip_hdr
 				if ( len > gre_len + 14 )
 					{
 					eth_len = 14;
-					proto_typ = ntohs(*((uint16*)(data + gre_len + eth_len - 2)));
+					proto_typ = ntohs(*((uint16_t*)(data + gre_len + eth_len - 2)));
 					}
 				else
 					{
@@ -533,7 +453,7 @@ void NetSessions::DoNextPacket(double t, const Packet* pkt, const IP_Hdr* ip_hdr
 					{
 					erspan_len = 8;
 					eth_len = 14;
-					proto_typ = ntohs(*((uint16*)(data + gre_len + erspan_len + eth_len - 2)));
+					proto_typ = ntohs(*((uint16_t*)(data + gre_len + erspan_len + eth_len - 2)));
 					}
 				else
 					{
@@ -550,7 +470,7 @@ void NetSessions::DoNextPacket(double t, const Packet* pkt, const IP_Hdr* ip_hdr
 					erspan_len = 12;
 					eth_len = 14;
 
-					auto flags = data + erspan_len - 1;
+					auto flags = data + gre_len + erspan_len - 1;
 					bool have_opt_header = ((*flags & 0x01) == 0x01);
 
 					if ( have_opt_header  )
@@ -564,7 +484,7 @@ void NetSessions::DoNextPacket(double t, const Packet* pkt, const IP_Hdr* ip_hdr
 							}
 						}
 
-					proto_typ = ntohs(*((uint16*)(data + gre_len + erspan_len + eth_len - 2)));
+					proto_typ = ntohs(*((uint16_t*)(data + gre_len + erspan_len + eth_len - 2)));
 					}
 				else
 					{
@@ -622,7 +542,7 @@ void NetSessions::DoNextPacket(double t, const Packet* pkt, const IP_Hdr* ip_hdr
 
 		if ( gre_version == 1 )
 			{
-			uint16 ppp_proto = ntohs(*((uint16*)(data + gre_len + 2)));
+			uint16_t ppp_proto = ntohs(*((uint16_t*)(data + gre_len + 2)));
 
 			if ( ppp_proto != 0x0021 && ppp_proto != 0x0057 )
 				{
@@ -719,53 +639,41 @@ void NetSessions::DoNextPacket(double t, const Packet* pkt, const IP_Hdr* ip_hdr
 		return;
 	}
 
-	HashKey* h = BuildConnIDHashKey(id);
-	if ( ! h )
-		reporter->InternalError("hash computation failed");
-
-	Connection* conn = 0;
+	ConnIDKey key = BuildConnIDKey(id);
+	Connection* conn = nullptr;
 
 	// FIXME: The following is getting pretty complex. Need to split up
 	// into separate functions.
-	conn = (Connection*) d->Lookup(h);
+	auto it = d->find(key);
+	if ( it != d->end() )
+		conn = it->second;
+
 	if ( ! conn )
 		{
-		conn = NewConn(h, t, &id, data, proto, ip_hdr->FlowLabel(), pkt, encapsulation);
+		conn = NewConn(key, t, &id, data, proto, ip_hdr->FlowLabel(), pkt, encapsulation);
 		if ( conn )
-			d->Insert(h, conn);
+			InsertConnection(d, key, conn);
 		}
 	else
 		{
 		// We already know that connection.
-		int consistent = CheckConnectionTag(conn);
-		if ( consistent < 0 )
+		if ( conn->IsReuse(t, data) )
 			{
-			delete h;
-			return;
-			}
-
-		if ( ! consistent || conn->IsReuse(t, data) )
-			{
-			if ( consistent )
-				conn->Event(connection_reused, 0);
+			conn->Event(connection_reused, 0);
 
 			Remove(conn);
-			conn = NewConn(h, t, &id, data, proto, ip_hdr->FlowLabel(), pkt, encapsulation);
+			conn = NewConn(key, t, &id, data, proto, ip_hdr->FlowLabel(), pkt, encapsulation);
 			if ( conn )
-				d->Insert(h, conn);
+				InsertConnection(d, key, conn);
 			}
 		else
 			{
-			delete h;
 			conn->CheckEncapsulation(encapsulation);
 			}
 		}
 
 	if ( ! conn )
-		{
-		delete h;
 		return;
-		}
 
 	int record_packet = 1;	// whether to record the packet at all
 	int record_content = 1;	// whether to record its data
@@ -813,7 +721,7 @@ void NetSessions::DoNextInnerPacket(double t, const Packet* pkt,
 		const IP_Hdr* inner, const EncapsulationStack* prev,
 		const EncapsulatingConn& ec)
 	{
-	uint32 caplen, len;
+	uint32_t caplen, len;
 	caplen = len = inner->TotalLen();
 
 	pkt_timeval ts;
@@ -887,16 +795,16 @@ int NetSessions::ParseIPPacket(int caplen, const u_char* const pkt, int proto,
 		return -1;
 		}
 
-	if ( (uint32)caplen != inner->TotalLen() )
-		return (uint32)caplen < inner->TotalLen() ? -1 : 1;
+	if ( (uint32_t)caplen != inner->TotalLen() )
+		return (uint32_t)caplen < inner->TotalLen() ? -1 : 1;
 
 	return 0;
 	}
 
-bool NetSessions::CheckHeaderTrunc(int proto, uint32 len, uint32 caplen,
+bool NetSessions::CheckHeaderTrunc(int proto, uint32_t len, uint32_t caplen,
                                    const Packet* p, const EncapsulationStack* encap)
 	{
-	uint32 min_hdr_len = 0;
+	uint32_t min_hdr_len = 0;
 	switch ( proto ) {
 	case IPPROTO_TCP:
 		min_hdr_len = sizeof(struct tcphdr);
@@ -942,29 +850,24 @@ bool NetSessions::CheckHeaderTrunc(int proto, uint32 len, uint32 caplen,
 FragReassembler* NetSessions::NextFragment(double t, const IP_Hdr* ip,
 					const u_char* pkt)
 	{
-	uint32 frag_id = ip->ID();
+	uint32_t frag_id = ip->ID();
 
-	ListVal* key = new ListVal(TYPE_ANY);
-	key->Append(new AddrVal(ip->SrcAddr()));
-	key->Append(new AddrVal(ip->DstAddr()));
-	key->Append(val_mgr->GetCount(frag_id));
+	FragReassemblerKey key = std::make_tuple(ip->SrcAddr(), ip->DstAddr(), frag_id);
 
-	HashKey* h = ch->ComputeHash(key, 1);
-	if ( ! h )
-		reporter->InternalError("hash computation failed");
+	FragReassembler* f = nullptr;
+	auto it = fragments.find(key);
+	if ( it != fragments.end() )
+		f = it->second;
 
-	FragReassembler* f = fragments.Lookup(h);
 	if ( ! f )
 		{
-		f = new FragReassembler(this, ip, pkt, h, t);
-		fragments.Insert(h, f);
-		Unref(key);
+		f = new FragReassembler(this, ip, pkt, key, t);
+		fragments[key] = f;
+		if ( fragments.size() > stats.max_fragments )
+			stats.max_fragments = fragments.size();
 		return f;
 		}
 
-	delete h;
-	Unref(key);
-
 	f->AddFragment(t, ip, pkt);
 	return f;
 	}
@@ -1020,11 +923,8 @@ Connection* NetSessions::FindConnection(Val* v)
 
 	id.is_one_way = 0;	// ### incorrect for ICMP connections
 
-	HashKey* h = BuildConnIDHashKey(id);
-	if ( ! h )
-		reporter->InternalError("hash computation failed");
-
-	Dictionary* d;
+	ConnIDKey key = BuildConnIDKey(id);
+	ConnectionMap* d;
 
 	if ( orig_portv->IsTCP() )
 		d = &tcp_conns;
@@ -1037,22 +937,22 @@ Connection* NetSessions::FindConnection(Val* v)
 		// This can happen due to pseudo-connections we
 		// construct, for example for packet headers embedded
 		// in ICMPs.
-		delete h;
 		return 0;
 		}
 
-	Connection* conn = (Connection*) d->Lookup(h);
-
-	delete h;
+	Connection* conn = nullptr;
+	auto it = d->find(key);
+	if ( it != d->end() )
+		conn = it->second;
 
 	return conn;
 	}
 
 void NetSessions::Remove(Connection* c)
 	{
-	HashKey* k = c->Key();
-	if ( k )
+	if ( c->IsKeyValid() )
 		{
+		const ConnIDKey& key = c->Key();
 		c->CancelTimers();
 
 		if ( c->ConnTransport() == TRANSPORT_TCP )
@@ -1066,9 +966,7 @@ void NetSessions::Remove(Connection* c)
 			}
 
 		c->Done();
-
-		if ( connection_state_remove )
-			c->Event(connection_state_remove, 0);
+		c->RemovalEvent();
 
 		// Zero out c's copy of the key, so that if c has been Ref()'d
 		// up, we know on a future call to Remove() that it's no
@@ -1077,17 +975,17 @@ void NetSessions::Remove(Connection* c)
 
 		switch ( c->ConnTransport() ) {
 		case TRANSPORT_TCP:
-			if ( ! tcp_conns.RemoveEntry(k) )
+			if ( tcp_conns.erase(key) == 0 )
 				reporter->InternalWarning("connection missing");
 			break;
 
 		case TRANSPORT_UDP:
-			if ( ! udp_conns.RemoveEntry(k) )
+			if ( udp_conns.erase(key) == 0 )
 				reporter->InternalWarning("connection missing");
 			break;
 
 		case TRANSPORT_ICMP:
-			if ( ! icmp_conns.RemoveEntry(k) )
+			if ( icmp_conns.erase(key) == 0 )
 				reporter->InternalWarning("connection missing");
 			break;
 
@@ -1097,7 +995,6 @@ void NetSessions::Remove(Connection* c)
 		}
 
 		Unref(c);
-		delete k;
 		}
 	}
 
@@ -1106,42 +1003,38 @@ void NetSessions::Remove(FragReassembler* f)
 	if ( ! f )
 		return;
 
-	HashKey* k = f->Key();
-
-	if ( k )
-		{
-		if ( ! fragments.RemoveEntry(k) )
-			reporter->InternalWarning("fragment reassembler not in dict");
-		}
-	else
-		reporter->InternalWarning("missing fragment reassembler hash key");
+	if ( fragments.erase(f->Key()) == 0 )
+		reporter->InternalWarning("fragment reassembler not in dict");
 
 	Unref(f);
 	}
 
 void NetSessions::Insert(Connection* c)
 	{
-	assert(c->Key());
+	assert(c->IsKeyValid());
 
-	Connection* old = 0;
+	Connection* old = nullptr;
 
 	switch ( c->ConnTransport() ) {
-	// Remove first. Otherwise the dictioanry would still
-	// reference the old key for already existing connections.
+	// Remove first. Otherwise the map would still reference the old key for
+	// already existing connections.
 
 	case TRANSPORT_TCP:
-		old = (Connection*) tcp_conns.Remove(c->Key());
-		tcp_conns.Insert(c->Key(), c);
+		old = LookupConn(tcp_conns, c->Key());
+		tcp_conns.erase(c->Key());
+		InsertConnection(&tcp_conns, c->Key(), c);
 		break;
 
 	case TRANSPORT_UDP:
-		old = (Connection*) udp_conns.Remove(c->Key());
-		udp_conns.Insert(c->Key(), c);
+		old = LookupConn(udp_conns, c->Key());
+		udp_conns.erase(c->Key());
+		InsertConnection(&udp_conns, c->Key(), c);
 		break;
 
 	case TRANSPORT_ICMP:
-		old = (Connection*) icmp_conns.Remove(c->Key());
-		icmp_conns.Insert(c->Key(), c);
+		old = LookupConn(icmp_conns, c->Key());
+		icmp_conns.erase(c->Key());
+		InsertConnection(&icmp_conns, c->Key(), c);
 		break;
 
 	default:
@@ -1155,7 +1048,6 @@ void NetSessions::Insert(Connection* c)
 		// Some clean-ups similar to those in Remove() (but invisible
 		// to the script layer).
 		old->CancelTimers();
-		delete old->Key();
 		old->ClearKey();
 		Unref(old);
 		}
@@ -1163,55 +1055,47 @@ void NetSessions::Insert(Connection* c)
 
 void NetSessions::Drain()
 	{
-	IterCookie* cookie = tcp_conns.InitForIteration();
-	Connection* tc;
-
-	while ( (tc = tcp_conns.NextEntry(cookie)) )
+	for ( const auto& entry : tcp_conns )
 		{
+		Connection* tc = entry.second;
 		tc->Done();
-		tc->Event(connection_state_remove, 0);
+		tc->RemovalEvent();
 		}
 
-	cookie = udp_conns.InitForIteration();
-	Connection* uc;
-
-	while ( (uc = udp_conns.NextEntry(cookie)) )
+	for ( const auto& entry : udp_conns )
 		{
+		Connection* uc = entry.second;
 		uc->Done();
-		uc->Event(connection_state_remove, 0);
+		uc->RemovalEvent();
 		}
 
-	cookie = icmp_conns.InitForIteration();
-	Connection* ic;
-
-	while ( (ic = icmp_conns.NextEntry(cookie)) )
+	for ( const auto& entry : icmp_conns )
 		{
+		Connection* ic = entry.second;
 		ic->Done();
-		ic->Event(connection_state_remove, 0);
+		ic->RemovalEvent();
 		}
-
-	ExpireTimerMgrs();
 	}
 
 void NetSessions::GetStats(SessionStats& s) const
 	{
-	s.num_TCP_conns = tcp_conns.Length();
-	s.cumulative_TCP_conns = tcp_conns.NumCumulativeInserts();
-	s.num_UDP_conns = udp_conns.Length();
-	s.cumulative_UDP_conns = udp_conns.NumCumulativeInserts();
-	s.num_ICMP_conns = icmp_conns.Length();
-	s.cumulative_ICMP_conns = icmp_conns.NumCumulativeInserts();
-	s.num_fragments = fragments.Length();
+	s.num_TCP_conns = tcp_conns.size();
+	s.cumulative_TCP_conns = stats.cumulative_TCP_conns;
+	s.num_UDP_conns = udp_conns.size();
+	s.cumulative_UDP_conns = stats.cumulative_UDP_conns;
+	s.num_ICMP_conns = icmp_conns.size();
+	s.cumulative_ICMP_conns = stats.cumulative_ICMP_conns;
+	s.num_fragments = fragments.size();
 	s.num_packets = num_packets_processed;
 
-	s.max_TCP_conns = tcp_conns.MaxLength();
-	s.max_UDP_conns = udp_conns.MaxLength();
-	s.max_ICMP_conns = icmp_conns.MaxLength();
-	s.max_fragments = fragments.MaxLength();
+	s.max_TCP_conns = stats.max_TCP_conns;
+	s.max_UDP_conns = stats.max_UDP_conns;
+	s.max_ICMP_conns = stats.max_ICMP_conns;
+	s.max_fragments = stats.max_fragments;
 	}
 
-Connection* NetSessions::NewConn(HashKey* k, double t, const ConnID* id,
-					const u_char* data, int proto, uint32 flow_label,
+Connection* NetSessions::NewConn(const ConnIDKey& k, double t, const ConnID* id,
+					const u_char* data, int proto, uint32_t flow_label,
 					const Packet* pkt, const EncapsulationStack* encapsulation)
 	{
 	// FIXME: This should be cleaned up a bit, it's too protocol-specific.
@@ -1264,29 +1148,22 @@ Connection* NetSessions::NewConn(HashKey* k, double t, const ConnID* id,
 		return 0;
 		}
 
-	bool external = conn->IsExternal();
-
-	if ( external )
-		conn->AppendAddl(fmt("tag=%s",
-					conn->GetTimerMgr()->GetTag().c_str()));
-
 	if ( new_connection )
-		{
 		conn->Event(new_connection, 0);
 
-		if ( external && connection_external )
-			{
-			conn->ConnectionEventFast(connection_external, 0, {
-				conn->BuildConnVal(),
-				new StringVal(conn->GetTimerMgr()->GetTag().c_str()),
-			});
-			}
-		}
-
 	return conn;
 	}
 
-bool NetSessions::IsLikelyServerPort(uint32 port, TransportProto proto) const
+Connection* NetSessions::LookupConn(const ConnectionMap& conns, const ConnIDKey& key)
+	{
+	auto it = conns.find(key);
+	if ( it != conns.end() )
+		return it->second;
+
+	return nullptr;
+	}
+
+bool NetSessions::IsLikelyServerPort(uint32_t port, TransportProto proto) const
 	{
 	// We keep a cached in-core version of the table to speed up the lookup.
 	static set<bro_uint_t> port_cache;
@@ -1313,9 +1190,9 @@ bool NetSessions::IsLikelyServerPort(uint32 port, TransportProto proto) const
 	return port_cache.find(port) != port_cache.end();
 	}
 
-bool NetSessions::WantConnection(uint16 src_port, uint16 dst_port,
+bool NetSessions::WantConnection(uint16_t src_port, uint16_t dst_port,
 					TransportProto transport_proto,
-					uint8 tcp_flags, bool& flip_roles)
+					uint8_t tcp_flags, bool& flip_roles)
 	{
 	flip_roles = false;
 
@@ -1360,45 +1237,6 @@ bool NetSessions::WantConnection(uint16 src_port, uint16 dst_port,
 	return true;
 	}
 
-TimerMgr* NetSessions::LookupTimerMgr(const TimerMgr::Tag* tag, bool create)
-	{
-	if ( ! tag )
-		{
-		DBG_LOG(DBG_TM, "no tag, using global timer mgr %p", timer_mgr);
-		return timer_mgr;
-		}
-
-	TimerMgrMap::iterator i = timer_mgrs.find(*tag);
-	if ( i != timer_mgrs.end() )
-		{
-		DBG_LOG(DBG_TM, "tag %s, using non-global timer mgr %p", tag->c_str(), i->second);
-		return i->second;
-		}
-	else
-		{
-		if ( ! create )
-			return 0;
-
-		// Create new queue for tag.
-		TimerMgr* mgr = new CQ_TimerMgr(*tag);
-		DBG_LOG(DBG_TM, "tag %s, creating new non-global timer mgr %p", tag->c_str(), mgr);
-		timer_mgrs.insert(TimerMgrMap::value_type(*tag, mgr));
-		double t = timer_mgr->Time() + timer_mgr_inactivity_timeout;
-		timer_mgr->Add(new TimerMgrExpireTimer(t, mgr));
-		return mgr;
-		}
-	}
-
-void NetSessions::ExpireTimerMgrs()
-	{
-	for ( TimerMgrMap::iterator i = timer_mgrs.begin();
-	      i != timer_mgrs.end(); ++i )
-		{
-		i->second->Expire();
-		delete i->second;
-		}
-	}
-
 void NetSessions::DumpPacket(const Packet *pkt, int len)
 	{
 	if ( ! pkt_dumper )
@@ -1406,7 +1244,7 @@ void NetSessions::DumpPacket(const Packet *pkt, int len)
 
 	if ( len != 0 )
 		{
-		if ( (uint32)len > pkt->cap_len )
+		if ( (uint32_t)len > pkt->cap_len )
 			reporter->Warning("bad modified caplen");
 		else
 			const_cast<Packet *>(pkt)->cap_len = len;
@@ -1445,23 +1283,14 @@ unsigned int NetSessions::ConnectionMemoryUsage()
 		// Connections have been flushed already.
 		return 0;
 
-	IterCookie* cookie = tcp_conns.InitForIteration();
-	Connection* tc;
+	for ( const auto& entry : tcp_conns )
+		mem += entry.second->MemoryAllocation();
 
-	while ( (tc = tcp_conns.NextEntry(cookie)) )
-		mem += tc->MemoryAllocation();
+	for ( const auto& entry : udp_conns )
+		mem += entry.second->MemoryAllocation();
 
-	cookie = udp_conns.InitForIteration();
-	Connection* uc;
-
-	while ( (uc = udp_conns.NextEntry(cookie)) )
-		mem += uc->MemoryAllocation();
-
-	cookie = icmp_conns.InitForIteration();
-	Connection* ic;
-
-	while ( (ic = icmp_conns.NextEntry(cookie)) )
-		mem += ic->MemoryAllocation();
+	for ( const auto& entry : icmp_conns )
+		mem += entry.second->MemoryAllocation();
 
 	return mem;
 	}
@@ -1474,23 +1303,14 @@ unsigned int NetSessions::ConnectionMemoryUsageConnVals()
 		// Connections have been flushed already.
 		return 0;
 
-	IterCookie* cookie = tcp_conns.InitForIteration();
-	Connection* tc;
+	for ( const auto& entry : tcp_conns )
+		mem += entry.second->MemoryAllocationConnVal();
 
-	while ( (tc = tcp_conns.NextEntry(cookie)) )
-		mem += tc->MemoryAllocationConnVal();
+	for ( const auto& entry : udp_conns )
+		mem += entry.second->MemoryAllocationConnVal();
 
-	cookie = udp_conns.InitForIteration();
-	Connection* uc;
-
-	while ( (uc = udp_conns.NextEntry(cookie)) )
-		mem += uc->MemoryAllocationConnVal();
-
-	cookie = icmp_conns.InitForIteration();
-	Connection* ic;
-
-	while ( (ic = icmp_conns.NextEntry(cookie)) )
-		mem += ic->MemoryAllocationConnVal();
+	for ( const auto& entry : icmp_conns )
+		mem += entry.second->MemoryAllocationConnVal();
 
 	return mem;
 	}
@@ -1503,17 +1323,35 @@ unsigned int NetSessions::MemoryAllocation()
 
 	return ConnectionMemoryUsage()
 		+ padded_sizeof(*this)
-		+ ch->MemoryAllocation()
-		// must take care we don't count the HaskKeys twice.
-		+ tcp_conns.MemoryAllocation() - padded_sizeof(tcp_conns) -
-		// 12 is sizeof(Key) from ConnID::BuildConnKey();
-		// it can't be (easily) accessed here. :-(
-			(tcp_conns.Length() * pad_size(12))
-		+ udp_conns.MemoryAllocation() - padded_sizeof(udp_conns) -
-			(udp_conns.Length() * pad_size(12))
-		+ icmp_conns.MemoryAllocation() - padded_sizeof(icmp_conns) -
-			(icmp_conns.Length() * pad_size(12))
-		+ fragments.MemoryAllocation() - padded_sizeof(fragments)
+		+ (tcp_conns.size() * (sizeof(ConnectionMap::key_type) + sizeof(ConnectionMap::value_type)))
+		+ (udp_conns.size() * (sizeof(ConnectionMap::key_type) + sizeof(ConnectionMap::value_type)))
+		+ (icmp_conns.size() * (sizeof(ConnectionMap::key_type) + sizeof(ConnectionMap::value_type)))
+		+ (fragments.size() * (sizeof(FragmentMap::key_type) + sizeof(FragmentMap::value_type)))
 		// FIXME: MemoryAllocation() not implemented for rest.
 		;
 	}
+
+void NetSessions::InsertConnection(ConnectionMap* m, const ConnIDKey& key, Connection* conn)
+	{
+	(*m)[key] = conn;
+
+	switch ( conn->ConnTransport() )
+		{
+		case TRANSPORT_TCP:
+			stats.cumulative_TCP_conns++;
+			if ( m->size() > stats.max_TCP_conns )
+				stats.max_TCP_conns = m->size();
+			break;
+		case TRANSPORT_UDP:
+			stats.cumulative_UDP_conns++;
+			if ( m->size() > stats.max_UDP_conns )
+				stats.max_UDP_conns = m->size();
+			break;
+		case TRANSPORT_ICMP:
+			stats.cumulative_ICMP_conns++;
+			if ( m->size() > stats.max_ICMP_conns )
+				stats.max_ICMP_conns = m->size();
+			break;
+		default: break;
+		}
+	}
